"""Implements a tag which handles manually stored data.

This demonstrates three patterns:

    * The builtin NodeData.Read/Write/CopyTo to store custom data over scene loading and node
      copying boundaries.
    * Caching data within a running Cinema 4D instance to share data over reallocation boundaries
      or similar conditions.
    * Just store things in the data container.
"""
import c4d
import random

class ExampleTagData(c4d.plugins.TagData):
    """Implements a tag which handles manually stored data.
    """
    # The plugin ID for the hook.
    ID_PLUGIN: int = 1061096

    # The class bound cache of parameters hashed over node UUIDs. The node UUID is MAXON_CREATOR_ID,
    # i.e., it expresses "sameness" of nodes over (mac_id, time, secret_sauce).
    # 
    # We could also hash things in other ways, for example the state of a set of parameters, the 
    # data container, or if we had tea or coffee for breakfast. It depends on which nodes you would 
    # consider "the same" in the context of what you want to cache.
    UUID_PARAMETER_CACHE: dict[bytes, list[int]] = {}

    @classmethod
    def Register(cls) -> None:
        """Registers the plugin hook.
        """
        if not c4d.plugins.RegisterTagPlugin(
                id=cls.ID_PLUGIN,
                str="Example Tag",
                info=c4d.TAG_EXPRESSION | c4d.TAG_VISIBLE,
                g=cls, 
                description="texample",
                icon=c4d.bitmaps.InitResourceBitmap(c4d.Tdisplay)):
            print(f"Warning: Failed to register '{cls}' tag plugin.")

    @staticmethod
    def GetNodeUuid(node: c4d.GeListNode) -> bytes:
        """Identifies a node over reallocation and scene reloading boundaries.

        The returned UUID will express what a user would consider one node. When a user creates
        a document, material, shader, object, tag, etc., it will get such UUID assigned and it
        will stay the same even over reallocation and scene reloading boundaries.
        """
        data: memoryview = node.FindUniqueID(c4d.MAXON_CREATOR_ID)
        if not isinstance(data, memoryview):
            raise RuntimeError("Could not access node UUID.")

        return bytes(data or None)
        
    def Init(self, node: c4d.GeListNode) -> bool:
        """Called to initialize a tag instance.

        Args:
            node: The BaseTag instance representing this plugin object.
        """
        self.InitAttr(node, bool, c4d.ID_MYTAG_SOMEPARAM)
        node[c4d.ID_MYTAG_SOMEPARAM] = False

        # Hash the node over its UUID to lookup if we already stored data under this hash. With the
        # `2023.2` API it is not possible skip the data init of dummy nodes, in the next major API
        # release it will be.
        uuid: bytes = ExampleTagData.GetNodeUuid(node)
        data: list | None = ExampleTagData.UUID_PARAMETER_CACHE.get(uuid, None)
        if data is None:
            data = [random.randint(0, 9) for i in range(random.randint(3, 5))]
            ExampleTagData.UUID_PARAMETER_CACHE[uuid] = data
            print(f"ExampleTagData.Init: Created the data '{data}'.")
        else:
            print(f"ExampleTagData.Init: Restored the data '{data}'.")

        self._data: list[int] = data

        # The safest route is to store all data in the data container of the node. The parameter ID 
        # ID_MYTAG_CUSTOMDATA is just a eum-value defined in the header file of the plugin as slot
        # for the custom data. It is neither reflected in the res nor string file and therefore does
        # not appear in the GUI.
        bc: c4d.BaseContainer = c4d.BaseContainer(c4d.ID_MYTAG_CUSTOMDATA)
        for i, v in enumerate(data):
            bc[i] = v
        
        node[c4d.ID_MYTAG_CUSTOMDATA] = bc

        return True

    def CopyTo(self, dest: "ExampleTagData", snode: c4d.GeListNode, dnode: c4d.GeListNode, 
               flags: int, trn: c4d.AliasTrans) -> bool:
        """Called by Cinema 4D when a node is copied to let allow the implementation to copy data
        not stored in the data container of the node.
        """
        dest._data = self._data
        print(f"ExampleTagData.CopyTo: Copied data from `{self}` to {dest}.")
        return True

    def Write(self, node: c4d.GeListNode, hf: c4d.storage.HyperFile) -> bool:
        """Called by Cinema 4D when an object is saved with a document to manually serialize data
        not stored in its data container..

        This method defines the serialization form which must be deserialized in NodeData.Read.
        """
        # Write the data in the form [count, item_1, ..., item_count]
        count: int = len(self._data)
        if not hf.WriteInt32(count):
            raise IOError("Could not serialize {self.__class.__name__} node.")

        success: list[bool] = [hf.WriteInt32(v) for v in self._data]
        if not all(success):
            raise IOError("Could not serialize {self.__class.__name__} node.")

        print(f"ExampleTagData.Read: Serialized `{self._data}` for {node}.")
        return True

    def Read(self, node: c4d.GeListNode, hf: c4d.storage.HyperFile, level: int) -> bool:
        """Called by Cinema 4D when an object is loaded from a document to manually deserialize data
        not stored in its data container..

        What must be done here is defined by the serialization pattern chosen in NodeData:Write.
        """
        # Get the count written into the first bin and then read #count bins for the actual values.
        count: int = hf.ReadInt32()
        if not count:
            raise IOError("Could not deserialize {self.__class.__name__} node.")

        data: list[int] = [hf.ReadInt32() for _ in range(count)]
        if None in data:
            raise IOError("Could not deserialize {self.__class.__name__} node.")
        
        print(f"ExampleTagData.Read: Deserialized `{data}` for {node}.")
        self._data = data
        return True

    def Execute(self, tag: c4d.BaseTag, doc: c4d.documents.BaseDocument, op: c4d.BaseObject,
                bt: c4d.threading.BaseThread, priority: int, flags: int) -> int:
        """Called when expressions are evaluated to let a tag modify a scene.

        Not used in this case.
        """
        print(f"ExampleTagData.Execute: "
              f"{', '.join([str(v) for _, v in tag[c4d.ID_MYTAG_CUSTOMDATA]])}")
        return c4d.EXECUTIONRESULT_OK

if __name__ == "__main__":
    ExampleTagData.Register()